import os
import sys
import re
import time
from datetime import datetime
from aqt.qt import *
import aqt
import anki.notes

from . import config
from .note_dialog import NoteDialog
from .progress_window import ProgressWindow
from .mi_japanese_handler import MiJHandler


addon_path = os.path.dirname(__file__)


class KindleDataModel(QSqlRelationalTableModel):

    def data(self, index, role=Qt.DisplayRole):

        if role == Qt.DisplayRole and index.column() == 6:
            data_val = super(KindleDataModel, self).data(index, role)
            if not data_val:
                return 'NONE'
            try:
                timestamp = int(data_val)
            except ValueError:
                return 'INVALID'
            date = datetime.fromtimestamp(timestamp/1000)
            date_str = date.strftime('%Y-%m-%d %H:%M:%S')
            return date_str
        
        return super(KindleDataModel, self).data(index, role)

    def resetFilter(self):
        super(KindleDataModel, self).setFilter("LOOKUPS.id not in (SELECT id FROM mdb.HANDLED_LOOKUPS)")

    def setFilter(self, filter_text, book_key=None):
        filter_text = filter_text.strip()

        filters = ['LOOKUPS.id not in (SELECT id FROM mdb.HANDLED_LOOKUPS)']
        if filter_text:
            filters.append("(usage LIKE '%%%s%%') or (word LIKE '%%%s%%')" % (filter_text, filter_text))
        if book_key:
            filters.append("book_key LIKE '%s'" % book_key)

        filter_str = ''
        if len(filters):
            filter_str = '(' + ') and ('.join(filters) + ')'            

        super(KindleDataModel, self).setFilter(filter_str)


class KindleWindow(QMainWindow):

    current_instance = None

    def __init__(self, parent, db_path):
        super(KindleWindow, self).__init__(parent)

        if KindleWindow.current_instance:
            raise ValueError('KindleWindow duplicate instance requested')
        KindleWindow.current_instance = self

        self.jhandler = MiJHandler(aqt.mw)

        self.open_db(db_path)

        self.setWindowTitle('Migaku Kindle')
        self.setWindowIcon(QIcon(os.path.join(addon_path, 'migaku.png')))

        self.setMinimumWidth(600)
        self.setMinimumHeight(400)

        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)
        self.lyt = QVBoxLayout(self.central_widget)

        self.search_lyt = QHBoxLayout()
        self.lyt.addLayout(self.search_lyt)

        self.search_line = QLineEdit()
        self.search_lyt.addWidget(self.search_line)
        self.book_box = QComboBox()
        self.book_box.setMaximumWidth(240)
        self.search_lyt.addWidget(self.book_box)
        self.search_button = QPushButton('Search')
        self.search_button.clicked.connect(self.on_search)
        self.search_lyt.addWidget(self.search_button)

                               # List all books that
        book_query = QSqlQuery('SELECT id, title FROM BOOK_INFO WHERE BOOK_INFO.id IN' \
                               # Have a lookup entry
                               '(SELECT DISTINCT book_key FROM LOOKUPS WHERE LOOKUPS.id not in ' \
                               # That isn't already imported
                               '(SELECT id FROM mdb.HANDLED_LOOKUPS));',
                               self.db)
        book_query.exec_()

        self.book_box.addItem('All Books')
        self.book_relation = [None]

        while book_query.next():
            self.book_box.addItem(book_query.value(1))
            self.book_relation.append(book_query.value(0))

        self.model = KindleDataModel(self, self.db)
        self.model.setTable('LOOKUPS')
        self.model.setRelation(1, QSqlRelation('WORDS', 'id', 'word'))
        self.model.setRelation(2, QSqlRelation('BOOK_INFO', 'id', 'title'))

        self.model.setHeaderData(1, Qt.Horizontal, 'Word')
        self.model.setHeaderData(2, Qt.Horizontal, 'Book')
        self.model.setHeaderData(5, Qt.Horizontal, 'Sentence')
        self.model.setHeaderData(6, Qt.Horizontal, 'Date')

        self.table_view = QTableView()
        self.table_view.setModel(self.model)
        self.table_view.verticalHeader().hide()
        self.table_view.horizontalHeader().hideSection(0)    # Hide id
        self.table_view.horizontalHeader().hideSection(3)    # Hide dict id
        self.table_view.horizontalHeader().hideSection(4)    # Hide pos
        self.table_view.horizontalHeader().moveSection(5, 2) # Move sentence before book title
        self.table_view.horizontalHeader().setSectionResizeMode(5, QHeaderView.Stretch) # Sentence fills remaining space
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_view.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.table_view.setSortingEnabled(True)
        self.table_view.setContextMenuPolicy(Qt.CustomContextMenu)

        self.model.resetFilter()
        self.model.select()
        self.table_view.resizeColumnsToContents()
        self.table_view.sortByColumn(6, Qt.DescendingOrder) # Sort by date, descending

        self.lyt.addWidget(self.table_view)

        self.actions_lyt = QHBoxLayout()
        self.lyt.addLayout(self.actions_lyt)

        self.deck_chooser_area = QWidget()
        self.deck_chooser = aqt.deckchooser.DeckChooser(aqt.mw, self.deck_chooser_area)
        self.actions_lyt.addWidget(self.deck_chooser_area)

        self.actions_lyt.addWidget(QLabel('Note'))
        self.note_btn = QPushButton(aqt.mw.col.models.current()['name'])
        self.actions_lyt.addWidget(self.note_btn)

        # Prevent deck button from expanding
        self.deck_chooser.deck.setSizePolicy(self.note_btn.sizePolicy())

        self.actions_lyt.addStretch()

        self.ja_checkbox = QCheckBox('Add Japanese Syntax')
        self.ja_checkbox.setChecked(config.get('ja_syntax', True))
        self.actions_lyt.addWidget(self.ja_checkbox)

        if not self.jhandler.activeNotes:
            self.ja_checkbox.hide()

        self.import_btn = QPushButton('Import')
        self.import_btn.clicked.connect(self.on_import_click)
        self.actions_lyt.addWidget(self.import_btn)

        # Connect signals/slots
        self.search_line.editingFinished.connect(self.on_search)
        self.table_view.doubleClicked.connect(self.on_table_double_click)
        self.table_view.customContextMenuRequested.connect(self.on_table_context_menu)
        del_shortcut = QShortcut(QKeySequence.Delete, self.table_view)
        del_shortcut.activated.connect(self.delete_selected)
        self.book_box.currentIndexChanged.connect(self.on_search)
        self.note_btn.clicked.connect(self.on_note_click)


    def closeEvent(self, event):
        self.close_db()
        event.accept()
        self.deck_chooser.cleanup()
        config.set('ja_syntax', self.ja_checkbox.isChecked(), do_write=True)
        KindleWindow.current_instance = None
        QMainWindow.closeEvent(self, event)


    def open_db(self, path):
        self.db_path = path

        # Open vocab.db
        self.db = QSqlDatabase.addDatabase('QSQLITE', 'kindle_vocab')
        self.db.setDatabaseName(self.db_path)
        self.db.open()

        # Attach imported words db
        mdb_path = os.path.join(aqt.mw.pm.profileFolder(), 'migaku_kindle.db')

        if not os.path.isfile(mdb_path):
            mdb = QSqlDatabase.addDatabase('QSQLITE', 'migaku_kindle')
            mdb.setDatabaseName(mdb_path)
            mdb.open()
            mdb.exec_('CREATE TABLE HANDLED_LOOKUPS (id TEXT PRIMARY KEY NOT NULL, reason INT)')
            mdb.close()
            QSqlDatabase.removeDatabase('migaku_kindle')

        self.db.exec_("ATTACH DATABASE '%s' AS mdb;" % mdb_path)

    
    def close_db(self):
        self.db.close()
        QSqlDatabase.removeDatabase('kindle_vocab')


    def on_search(self):
        filter_text = self.search_line.text().strip()

        book_key = None
        curr_book_box_idx = self.book_box.currentIndex()
        if curr_book_box_idx > 0:
            book_key = self.book_relation[curr_book_box_idx]

        self.model.setFilter(filter_text, book_key)
        self.model.select()


    def delete_selected(self):
        selected_rows = self.table_view.selectionModel().selectedRows()
        if len(selected_rows) < 1:
            return

        lookup_ids = []
        for index in selected_rows:
            lookup_ids.append(index.data())

        self.add_handled_lookups(lookup_ids, 0)
        self.model.select()

    
    def on_note_click(self):
        dlg = NoteDialog(self)
        dlg.exec_()
        self.note_btn.setText(aqt.mw.col.models.current()['name'])


    # reason: 0=Ignored, 1=Imported
    def add_handled_lookup(self, lookup_id, reason):
        self.db.exec_("INSERT INTO mdb.HANDLED_LOOKUPS (id, reason) VALUES ('%s', %d)" % (lookup_id, reason))

    def add_handled_lookups(self, lookup_ids, reason):
        if len(lookup_ids) < 1:
            return
        query_values = ','.join("('%s',%d)" % (i,reason) for i in lookup_ids)
        query = 'INSERT INTO mdb.HANDLED_LOOKUPS (id, reason) VALUES ' + query_values
        self.db.exec_(query)

    
    def on_import_click(self):
        selected_rows = self.table_view.selectionModel().selectedRows()

        if len(selected_rows) < 1:
            QMessageBox.information(self, 'Migaku Kindle', 'Please selected rows (Ctrl/Shift click for multi select) and then press "Import".')
            return

        deck_id = self.deck_chooser.selectedId()

        model = aqt.mw.col.models.current()
        model_id = str(model['id'])

        associations = config.get('notes').get(model_id, [None, None, None])

        if not any(associations):
            QMessageBox.information(self, 'Migaku Kindle', 'Export configuration not set for current note type.\n\nClick the "Note" button to configure it.')
            return

        for a in associations:
            if a:
                has_field = False
                for f in model['flds']:
                    if f['name'] == a:
                        has_field = True
                        break
                if not has_field:
                    QMessageBox.information(self, 'Migaku Kindle', 'Export configuration for current note type is invalid.\n\nClick the "Note" button to configure it.')
                    return

        use_dict = False
        if hasattr(aqt.mw, 'dictionaryInit'):
            dict_plugin_main = sys.modules[aqt.mw.dictionaryInit.__module__]
            dict_cfg = aqt.mw.addonManager.getConfig(dict_plugin_main.__name__)
            dict_fb = dict_cfg['frontBracket']
            dict_bb = dict_cfg['backBracket']
            dict_lang = dict_cfg['ForvoLanguage']
            dict_mapping = {}
            for d in aqt.mw.miDictDB.getAllDicts():
                dict_name = aqt.mw.miDictDB.cleanDictName(d)
                dict_mapping[dict_name] = d
            dict_data = config.get('notes_dict').get(model_id, [[], 1])
            dict_names = dict_data[0]
            dict_max_per_dict = dict_data[1]
            dict_dst = dict_data[2]
            has_dict_dst = False
            for f in model['flds']:
                if f['name'] == dict_dst:
                    has_dict_dst = True
                    break
            use_dict = len(dict_names) > 0 and has_dict_dst

        notes_num = len(selected_rows)
        show_progress = False
        start_time = time.time()
    
        aqt.mw.requireReset()

        for note_i, index in enumerate(selected_rows):
            
            if not show_progress and note_i > 0:
                now = time.time()
                delta = now - start_time
                progress = note_i/notes_num
                est_time = delta / progress
                if est_time > 3.0:
                    progress_window = ProgressWindow('Exporting Cards...', 'Migaku Kindle')
                    progress_window.set_value(progress)
                    progress_window.show()
                    aqt.mw.app.processEvents()
                    show_progress = True
            
            if show_progress and progress_window.was_closed:
                break                
            
            row = index.row()
            lookup_id = index.data()

            sentence = self.model.data(self.model.index(row, 5))
            word = self.model.data(self.model.index(row, 1))
            book = self.model.data(self.model.index(row, 2))

            note = anki.notes.Note(aqt.mw.col, model)

            if associations[0]:
                note[associations[0]] = sentence

            if associations[1]:
                note[associations[1]] = word

            if associations[2]:
                note[associations[2]] = book

            if self.ja_checkbox.isChecked():
                note = self.jhandler.attemptGenerate(note)

            if use_dict:
                term = re.sub(r'<[^>]+>', '', word) 
                term = re.sub(r'\[[^\]]+?\]', '', term)

                if term:
                    results = []

                    for dict_n in dict_names:
                        if dict_n == 'Google Images':
                            results.append(dict_plugin_main.exportGoogleImages(term, dict_max_per_dict))
                        elif dict_n == 'Forvo':
                            results.append(dict_plugin_main.exportForvoAudio(term, dict_max_per_dict, dict_lang))
                        elif dict_n in dict_mapping:
                            dresults, dh, th = aqt.mw.miDictDB.getDefForMassExp(term, dict_mapping[dict_n], str(dict_max_per_dict), dict_n)
                            results.append(dict_plugin_main.formatDefinitions(dresults, th, dh, dict_fb, dict_bb))

                    note[dict_dst] = '<br><br>'.join([i for i in results if i != ''])

            aqt.mw.col.add_note(note, deck_id)
            self.add_handled_lookup(lookup_id, 1)

            if show_progress:
                progress_window.set_value((note_i+1) / notes_num)
                aqt.mw.app.processEvents()

        self.model.select()
        aqt.mw.maybeReset()

        if show_progress:
            progress_window.close()
            progress_window.close()


    def on_table_double_click(self, index):
        row = index.row()
        word = self.model.data(self.model.index(row, 1))

        if hasattr(aqt.mw, 'dictionaryInit'):
            dict_plugin_main = sys.modules[aqt.mw.dictionaryInit.__module__]
            dict_plugin_main.searchTermList([word])

    
    def on_table_context_menu(self, point):

        if len(self.table_view.selectionModel().selectedRows()) > 1:
            return
        
        index = self.table_view.indexAt(point)

        if not index.isValid():
            return

        row = index.row()
        word = self.model.data(self.model.index(row, 1))
        sentence = self.model.data(self.model.index(row, 5))
        book = self.model.data(self.model.index(row, 2))

        menu = QMenu(self)

        action_lookup = None
        if hasattr(aqt.mw, 'dictionaryInit'):
            action_lookup = menu.addAction('Search in Dictionary')

        action_copy_word = menu.addAction('Copy Word')
        action_copy_sentence = menu.addAction('Copy Sentence')
        action_copy_book = menu.addAction('Copy Book')

        action = menu.exec_(self.table_view.mapToGlobal(point))

        if action == action_copy_word:
            QApplication.clipboard().setText(word)
        elif action == action_copy_sentence:
            QApplication.clipboard().setText(sentence)
        elif action == action_copy_book:
            QApplication.clipboard().setText(book)
        elif action_lookup and action == action_lookup:
            dict_plugin_main = sys.modules[aqt.mw.dictionaryInit.__module__]
            dict_plugin_main.searchTermList([word])
